# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Type hints for common dimod inputs.
"""
import typing

from typing import Collection, Hashable, Mapping, NamedTuple, Sequence, Tuple, Union

import numpy as np

from dimod.vartypes import VartypeLike

try:
    from numpy.typing import ArrayLike, DTypeLike
except ImportError:
    # support numpy < 1.20
    ArrayLike = typing.Sequence
    DTypeLike = typing.Any

try:
    from numpy.typing import NDArray
except ImportError:
    # support numpy < 1.21
    NDArray = typing.Sequence


__all__ = ['Bias',
           'GraphLike',
           'Polynomial',
           'SampleLike',
           'SamplesLike',
           'Variable',
           'VartypeLike',
           ]

# use float for python types, https://www.python.org/dev/peps/pep-0484/#the-numeric-tower
# exclude np.complexfloating from numpy types
Bias = Union[float, np.floating, np.integer]

Variable = Hashable  # todo: exclude None

try:
    import networkx as nx
except ImportError:
    GraphLike = Union[
        int,  # number of nodes
        Tuple[Collection[Variable], Collection[Tuple[Variable, Variable]]],  # nodes/edges
        Collection[Tuple[Variable, Variable]],  # edges
        ]
else:
    GraphLike = Union[
        int,  # number of nodes
        Tuple[Collection[Variable], Collection[Tuple[Variable, Variable]]],  # nodes/edges
        Collection[Tuple[Variable, Variable]],  # edges
        nx.Graph,
        ]

Polynomial = Mapping[Sequence[Variable], Bias]

class QuadraticVectors(NamedTuple):
    row_indices: NDArray[np.integer]
    col_indices: NDArray[np.integer]
    biases: NDArray[np.floating]


class BQMVectors(NamedTuple):
    linear_biases: NDArray[np.floating]
    quadratic: QuadraticVectors
    offset: Bias


class LabelledBQMVectors(NamedTuple):
    linear_biases: NDArray[np.floating]
    quadratic: QuadraticVectors
    offset: Bias
    labels: Sequence[Variable]


class DQMVectors(NamedTuple):
    case_starts: NDArray[np.integer]
    linear_biases: NDArray[np.floating]
    quadratic: QuadraticVectors
    labels: Sequence[Variable]
    offset: Bias


SampleLike = Union[
    typing.Sequence[float],
    typing.Mapping[Variable, Bias],
    ArrayLike,  # this is overgenerous, but we cannot easily specify it better
    ]


SamplesLike = Union[
    SampleLike,
    typing.Sequence[typing.Sequence[float]],  # 2d array
    typing.Tuple[typing.Sequence[float], typing.List[Variable]],
    typing.Tuple[typing.Sequence[typing.Sequence[float]], typing.List[Variable]],
    typing.Sequence[SampleLike],
    typing.Iterator[SampleLike],
    ]



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
A simple text encoding of dimod binary quadratic models (BQMs).

The COOrdinate_ list is a sparse matrix representation which can be used to
store BQMs. This format is best used when readability is important.

.. note:: This format works only for BQMs labelled with positive integers.

.. _COOrdinate: https://en.wikipedia.org/wiki/Sparse_matrix#Coordinate_list_(COO)

Examples:

    >>> from dimod.serialization import coo

    Serialize a QUBO.

    >>> Q = {(0, 0): -1, (0, 1): 1, (1, 2): -4.5}
    >>> bqm = dimod.BinaryQuadraticModel.from_qubo(Q)
    >>> print(coo.dumps(bqm))
    0 0 -1.000000
    0 1 1.000000
    1 2 -4.500000

    Include the :class:`~dimod.Vartype` as a header.

    >>> Q = {(0, 0): -1, (0, 1): 1, (1, 2): -4.5}
    >>> bqm = dimod.BinaryQuadraticModel.from_qubo(Q)
    >>> print(coo.dumps(bqm, vartype_header=True))
    # vartype=BINARY
    0 0 -1.000000
    0 1 1.000000
    1 2 -4.500000

    Load from a COO string. You must specify a :class:`~dimod.Vartype` if absent
    from the input string.

    >>> coo_string = '''
    ... 0 0 -1
    ... 0 1 1
    ... 1 2 -4.5
    ... '''
    >>> coo.loads(coo_string, vartype=dimod.BINARY)
    BinaryQuadraticModel({0: -1.0, 1: 0.0, 2: 0.0}, {(1, 0): 1.0, (2, 1): -4.5}, 0.0, 'BINARY')

    You can provide the :class:`~dimod.Vartype` as a header in the string.

    >>> coo_string = '''
    ... # vartype=BINARY
    ... 0 0 -1
    ... 0 1 1
    ... 1 2 -4.5
    ... '''
    >>> coo.loads(coo_string)
    BinaryQuadraticModel({0: -1.0, 1: 0.0, 2: 0.0}, {(1, 0): 1.0, (2, 1): -4.5}, 0.0, 'BINARY')

"""

import re
import typing
import warnings

from numbers import Integral

from dimod.vartypes import Vartype
from dimod.binary_quadratic_model import BinaryQuadraticModel

_LINE_REGEX = r'^\s*(\d+)\s+(\d+)\s+([+-]?\d*(?:\.\d+)?)\s*$'
"""
Each line should look like

0 1 2.000000

where 0, 1 are the variable labels and 2.0 is the bias.
"""

_VARTYPE_HEADER_REGEX = r'^[ \t\f]*#.*?vartype[:=][ \t]*([-_.a-zA-Z0-9]+)'
"""
The header should be in the first line and look like

# vartype=SPIN

"""


def dumps(bqm: BinaryQuadraticModel, vartype_header: bool = False) -> str:
    """Dump a binary quadratic model to a string in COOrdinate format.

    Args:

        bqm: Binary quadratic model to save as a string in COO format.

        vartype_header: If True, prefixes the :class:`~dimod.Vartype` to the output
            as a comment line (e.g., ``# vartype=SPIN``).

    """
    return '\n'.join(_iter_triplets(bqm, vartype_header))


def dump(bqm: BinaryQuadraticModel, fp: typing.TextIO, vartype_header: bool = False):
    """Dump a binary quadratic model to a file in COOrdinate format.

    Args:

        bqm: Binary quadratic model to save to a file in COO format.

        fp: File pointer to a file opened in write mode.

        vartype_header: Prefix the :class:`~dimod.Vartype` to the output.

    Examples:
        >>> from dimod.serialization import coo
        >>> bqm = dimod.BinaryQuadraticModel.from_ising({0: 1, 1: 2}, {(0, 1): -1})
        >>> with open('my_saved_bqm.txt', 'w') as f:           # doctest: +SKIP
        ...    coo.dump(bqm, f, vartype_header=True)

    """
    for triplet in _iter_triplets(bqm, vartype_header):
        fp.write('%s\n' % triplet)


def loads(s: str, cls: None = None,
          vartype: typing.Optional[Vartype] = None) -> BinaryQuadraticModel:
    """Load a binary quadratic model from a COOrdinate-formatted string.

    Args:

        s: String containing a COO-formatted binary quadratic model.

        cls: Deprecated. Does nothing.

        vartype: The valid variable types for binary quadratic models, is
          one of:

            * :class:`~dimod.Vartype.SPIN`, ``'SPIN'``, ``{-1, +1}``
            * :class:`~dimod.Vartype.BINARY`, ``'BINARY'``, ``{0, 1}``

    .. deprecated:: 0.10.15

        The ``cls`` keyword will be removed in dimod 0.12.0. It currently does
        nothing.
    """

    if cls is not None:
        warnings.warn("cls keyword argument is deprecated since 0.10.15 and will "
                      "be removed in dimod 0.11. Does nothing.", DeprecationWarning,
                      stacklevel=2)

    return load(s.split('\n'), vartype=vartype)


def load(fp: typing.TextIO, cls: None = None,
         vartype:  typing.Optional[Vartype] = None) -> BinaryQuadraticModel:
    """Load a binary quadratic model from a COOrdinate-formatted file.

    Args:

        fp: File pointer to a file opened in read mode.

        cls: Deprecated. Does nothing.

        vartype: The valid variable types for binary quadratic models, is
          one of:

            * :class:`~dimod.Vartype.SPIN`, ``'SPIN'``, ``{-1, +1}``
            * :class:`~dimod.Vartype.BINARY`, ``'BINARY'``, ``{0, 1}``

    Examples:
        >>> from dimod.serialization import coo
        >>> with open('my_bqm.txt', 'r') as f:          # doctest: +SKIP
        ...     my_bqm = coo.load(f, vartype=dimod.Vartype.SPIN)
    
    .. deprecated:: 0.10.15

        The ``cls`` keyword will be removed in dimod 0.12.0. It currently does
        nothing.

    """

    if cls is not None:
        warnings.warn("cls keyword argument is deprecated since 0.10.15 and will "
                      "be removed in dimod 0.11. Does nothing.", DeprecationWarning,
                      stacklevel=2)

    pattern = re.compile(_LINE_REGEX)
    vartype_pattern = re.compile(_VARTYPE_HEADER_REGEX)

    triplets = []
    for line in fp:
        triplets.extend(pattern.findall(line))

        vt = vartype_pattern.findall(line)
        if vt:
            if vartype is None:
                vartype = vt[0]
            else:
                if isinstance(vartype, str):
                    vartype = Vartype[vartype]
                else:
                    vartype = Vartype(vartype)
                if Vartype[vt[0]] != vartype:
                    raise ValueError("vartypes from headers and/or inputs do not match")

    if vartype is None:
        raise ValueError("vartype must be provided either as a header or as an argument")

    bqm = BinaryQuadraticModel.empty(vartype)

    for u, v, bias in triplets:
        if u == v:
            bqm.add_variable(int(u), float(bias))
        else:
            bqm.add_interaction(int(u), int(v), float(bias))

    return bqm


def _iter_triplets(bqm, vartype_header):
    if not all(isinstance(v, Integral) and v >= 0 for v in bqm.linear):
        raise ValueError("only positive index-labeled binary quadratic models can be dumped to COOrdinate format")

    if vartype_header:
        yield '# vartype=%s' % bqm.vartype.name

    # developer note: we could (for some threshold sparseness) sort the neighborhoods,
    # but this is simple and probably sufficient

    variables = sorted(bqm.variables)  # integer labeled so we can sort in py3

    for idx, u in enumerate(variables):
        for v in variables[idx:]:
            if u == v and bqm.linear[u]:
                yield '%d %d %f' % (u, u, bqm.linear[u])
            elif u in bqm.adj[v]:
                yield '%d %d %f' % (u, v, bqm.adj[u][v])



# Copyright 2020 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import collections.abc as abc

cimport cython

from cython.operator cimport preincrement as inc, dereference as deref
from cython.operator cimport preincrement as preinc  # for when it matters
from libcpp.algorithm cimport lower_bound, sort, unique
from libcpp.unordered_set cimport unordered_set

import numpy as np

from dimod.utilities import asintegerarrays, asnumericarrays
from dimod.cyutilities cimport as_numpy_float

BIAS_DTYPE = np.float64
INDEX_DTYPE = np.int32

# developer note: for whatever reason this is the only way we can sort a vector
# of structs in clang. If we do it in pure cython we get segmentation faults
cdef extern from *:
    """
    #include <vector>
    #include <algorithm>
    #include <cstdint>

    struct LinearTerm {
        std::int64_t variable;
        std::int64_t case_v;
        double bias;
    };

    bool term_lt(const LinearTerm& lhs, const LinearTerm& rhs) {
        return lhs.case_v < rhs.case_v;
    }

    void sort_terms(std::vector<LinearTerm>& terms) {
        std::sort(terms.begin(), terms.end(), term_lt);
    }
    """
    struct LinearTerm:
        np.int64_t variable
        np.int64_t case "case_v"
        double bias
    void sort_terms(vector[LinearTerm]&)


cdef class cyDiscreteQuadraticModel:

    def __init__(self):
        self.case_starts_.push_back(0)

        self.dtype = np.dtype(BIAS_DTYPE)
        self.case_dtype = np.dtype(INDEX_DTYPE)

    @property
    def adj(self):
        return self.adj_

    @property
    def offset(self):
        cdef bias_type offset = self.cppbqm.offset()
        return as_numpy_float(offset)

    @offset.setter
    def offset(self, bias_type offset):
        self._set_offset(offset)

    cdef void _add_offset(self, bias_type offset):
        cdef bias_type *off = &(self.cppbqm.offset())
        off[0] += offset

    cdef void _set_linear(self, Py_ssize_t vi, bias_type bias):
        # unsafe version of .set_linear
        cdef bias_type *b = &(self.cppbqm.linear(vi))
        b[0] = bias

    cdef void _set_offset(self, bias_type offset):
        cdef bias_type *off = &(self.cppbqm.offset())
        off[0] = offset

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def add_linear_equality_constraint(self, object terms,
                                       bias_type lagrange_multiplier, bias_type constant):
        # adjust energy offset
        self._add_offset(lagrange_multiplier * constant * constant)

        # resolve the terms from a python object into a C++ object
        cdef vector[LinearTerm] cppterms

        # can allocate it if we already know the size
        if isinstance(terms, abc.Sized):
            cppterms.reserve(len(terms))

        # put the generator or list into our C++ objects
        cdef Py_ssize_t v, case_v
        cdef bias_type bias
        cdef LinearTerm term
        for v, case_v, bias in terms:
            if case_v >= self.num_cases(v):  # also checks variable
                raise ValueError("case out of range")

            term.variable = v
            term.case = case_v + self.case_starts_[v]
            term.bias = bias
            cppterms.push_back(term)

        # sort and sum duplicates in terms
        sort_terms(cppterms)

        first = cppterms.begin()
        last = cppterms.end()
        if first != last:
            result = first

            while preinc(first) != last:
                if deref(result).case == deref(first).case:
                    # first is a duplicate
                    deref(result).bias = deref(result).bias + deref(first).bias
                else:
                    # advance result and set it to equal the current first
                    inc(result)
                    if result != first:
                        deref(result).variable = deref(first).variable
                        deref(result).case = deref(first).case
                        deref(result).bias = deref(first).bias

            # finally advance result and delete the rest of the vector
            cppterms.erase(preinc(result), cppterms.end())


        # add the biases to the BQM
        cdef Py_ssize_t num_terms = cppterms.size()

        cdef index_type u
        cdef Py_ssize_t i, j
        cdef index_type cu, cv
        cdef bias_type lbias, qbias, u_bias, v_bias
        for i in range(num_terms):
            u = cppterms[i].variable
            cu = cppterms[i].case
            u_bias = cppterms[i].bias

            lbias = lagrange_multiplier * u_bias * (2 * constant + u_bias)
            self._set_linear(cu, lbias + self.cppbqm.linear(cu))

            for j in range(i + 1, num_terms):
                cv = cppterms[j].case
                v_bias = cppterms[j].bias

                # interactions between cases within a variable never contribute
                # energy
                if u == cppterms[j].variable:
                    continue

                qbias = 2 * lagrange_multiplier * u_bias * v_bias

                self.cppbqm.add_quadratic(cu, cv, qbias)

        # now track the variables we used
        cdef unordered_set[index_type] variable_set
        for i in range(cppterms.size()):
            variable_set.insert(cppterms[i].variable)

        # sort the variables.
        cdef vector[index_type] variables
        variables.insert(variables.begin(), variable_set.begin(), variable_set.end())
        sort(variables.begin(), variables.end())

        # finally fix the adjacency
        cdef Py_ssize_t vi, ni
        for i in range(variables.size()):
            v = variables[i]

            # this is optimized for when variables are relatively large
            # compared adj (at 5000 variables the crossover point is %13)
            vit = variables.begin()
            nit = self.adj_[v].begin()
            while vit != variables.end():
                if deref(vit) == v:
                    inc(vit)
                elif nit == self.adj_[v].end():
                    nit = self.adj_[v].insert(nit, deref(vit))
                    inc(nit)
                    inc(vit)
                elif deref(vit) < deref(nit):
                    nit = self.adj_[v].insert(nit, deref(vit))
                    inc(nit)
                    inc(vit)
                elif deref(vit) > deref(nit):
                    inc(nit)
                else:  # *vit == *nit
                    inc(nit)
                    inc(vit)

    cpdef Py_ssize_t add_variable(self, Py_ssize_t num_cases) except -1:
        """Add a discrete variable.

        Args:
            num_cases (int): The number of cases in the variable.

        Returns:
            int: The label of the new variable.

        """

        if num_cases <= 0:
            raise ValueError("num_cases must be a positive integer")

        cdef index_type v = self.adj_.size()  # index of new variable

        self.adj_.resize(v+1)

        cdef Py_ssize_t i
        for i in range(num_cases):
            self.cppbqm.add_variable()

        self.case_starts_.push_back(self.cppbqm.num_variables())

        return v

    def copy(self):
        cdef cyDiscreteQuadraticModel dqm = type(self)()

        dqm.cppbqm = self.cppbqm
        dqm.case_starts_ = self.case_starts_
        dqm.adj_ = self.adj_
        dqm._set_offset(self.cppbqm.offset())

        return dqm

    def degree(self, Py_ssize_t vi):
        if not 0 <= vi < self.adj_.size():
            raise ValueError("not a variable")
        return self.adj_[vi].size()

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef bias_type[:] energies(self, index_type[:, :] samples):
        
        if samples.shape[1] != self.num_variables():
            raise ValueError("Given sample(s) have incorrect number of variables")

        cdef Py_ssize_t num_samples = samples.shape[0]
        cdef index_type num_variables = samples.shape[1]

        cdef bias_type[:] energies = np.full(num_samples, self.offset, dtype=self.dtype)

        cdef Py_ssize_t si, vi
        cdef index_type cu, case_u, cv, case_v
        cdef index_type u, v
        for si in range(num_samples):  # this could be parallelized
            for u in range(num_variables):
                case_u = samples[si, u]

                if case_u >= self.num_cases(u):
                    raise ValueError("invalid case")

                cu = self.case_starts_[u] + case_u

                energies[si] += self.cppbqm.linear(cu)

                for vi in range(self.adj_[u].size()):
                    v = self.adj_[u][vi]

                    # we only care about the lower triangle
                    if v > u:
                        break

                    case_v = samples[si, v]

                    cv = self.case_starts_[v] + case_v

                    energies[si] += self.cppbqm.quadratic(cu, cv)

        return energies

    @classmethod
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _from_numpy_vectors(cls,
                            Integral32plus[::1] case_starts,
                            Numeric32plus[::1] linear_biases,
                            Integral32plus[::1] irow,
                            Integral32plus[::1] icol,
                            Numeric32plus[::1] quadratic_biases,
                            bias_type offset):
        """Equivalent of from_numpy_vectors with fused types."""

        # some input checking
        # variable declarations we'll use throughout
        cdef Py_ssize_t u, v  # variables
        cdef Py_ssize_t ci, cj  # case indices
        cdef Py_ssize_t qi

        # constants
        cdef Py_ssize_t num_variables = case_starts.shape[0]
        cdef Py_ssize_t num_cases = linear_biases.shape[0]
        cdef Py_ssize_t num_interactions = irow.shape[0]

        # check that starts and linear_biases are correct and consistent with eachother
        for v in range(case_starts.shape[0] - 1):
            if case_starts[v+1] < case_starts[v]:
                raise ValueError("case_starts is not correctly ordered")

            if case_starts[v+1] >= linear_biases.shape[0]:
                raise ValueError("case_starts does not match linear_biases")

        # check that the quadratic are correct and consistent with eachother
        if not (irow.shape[0] == icol.shape[0] == quadratic_biases.shape[0]):
            raise ValueError("inconsistent lengths for irow, icol, qdata")
        for qi in range(irow.shape[0]):
            if not (0 <= irow[qi] < num_cases):
                raise ValueError("irow refers to case {} which is out of range"
                                 "".format(irow[qi]))
            if not (0 <= icol[qi] < num_cases):
                raise ValueError("icol refers to case {} which is out of range"
                                 "".format(icol[qi]))
            if irow[qi] == icol[qi]:
                raise ValueError("quadratic data contains a self-loop")

        cdef cyDiscreteQuadraticModel dqm = cls()

        # set the BQM
        if num_interactions:
            dqm.cppbqm.add_quadratic(
                &irow[0], &icol[0], &quadratic_biases[0],
                num_interactions)

        # add the linear biases
        while dqm.cppbqm.num_variables() < num_cases:
            dqm.cppbqm.add_variable()
        for ci in range(num_cases):
            dqm._set_linear(ci, linear_biases[ci])

        # set the case starts
        dqm.case_starts_.resize(case_starts.shape[0] + 1)
        for v in range(case_starts.shape[0]):
            dqm.case_starts_[v] = case_starts[v]
        dqm.case_starts_[case_starts.shape[0]] = dqm.cppbqm.num_variables()

        # and finally the adj. This is not really the memory bottleneck so
        # we can build an intermediate (unordered) set version
        cdef vector[unordered_set[index_type]] adjset
        adjset.resize(num_variables)
        u = 0
        for ci in range(dqm.cppbqm.num_variables()):
            
            # we've been careful so don't need ui < case_starts.size() - 1
            while ci >= dqm.case_starts_[u+1]:
                u += 1

            span = dqm.cppbqm.neighborhood(ci)

            v = 0
            while span.first != span.second:
                cj = deref(span.first).first

                # see above note
                while cj >= dqm.case_starts_[v+1]:
                    v += 1

                adjset[u].insert(v)

                inc(span.first)

        # now put adjset into adj
        dqm.adj_.resize(num_variables)
        for v in range(num_variables):
            dqm.adj_[v].insert(dqm.adj_[v].begin(),
                               adjset[v].begin(), adjset[v].end())
            sort(dqm.adj_[v].begin(), dqm.adj_[v].end())

        # do one last final check for self-loops within a variable
        for v in range(num_variables):
            for ci in range(dqm.case_starts_[v], dqm.case_starts_[v+1]):
                span2 = dqm.cppbqm.neighborhood(ci, dqm.case_starts_[v])
                if span2.first == span2.second:
                    continue
                if deref(span2.first).first < dqm.case_starts_[v+1]:
                    raise ValueError("A variable has a self-loop")

        # add provided offset to dqm
        dqm._set_offset(offset)

        return dqm

    @classmethod
    def from_numpy_vectors(cls, case_starts, linear_biases, quadratic, offset = 0):

        cdef cyDiscreteQuadraticModel obj = cls()

        try:
            irow, icol, quadratic_biases = quadratic
        except ValueError:
            raise ValueError("quadratic should be a 3-tuple")

        # We need:
        # * numpy ndarrays
        # * contiguous memory
        # * irow.dtype == icol.dtype == case_starts.dtype
        # * ldata.dtype==qdata.dtype
        # * 32 or 64 bit dtypes
        case_starts, icol, irow = asintegerarrays(
            case_starts, icol, irow, min_itemsize=4, requirements='C')
        ldata, qdata = asnumericarrays(
            linear_biases, quadratic_biases, min_itemsize=4, requirements='C')

        return cls._from_numpy_vectors(case_starts, ldata, irow, icol, qdata, offset)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def get_linear(self, index_type v):

        cdef Py_ssize_t num_cases = self.num_cases(v)

        biases = np.empty(num_cases, dtype=np.float64)
        cdef bias_type[:] biases_view = biases

        cdef Py_ssize_t c
        for c in range(num_cases):
            biases_view[c] = self.cppbqm.linear(self.case_starts_[v] + c)

        return biases

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef bias_type get_linear_case(self, index_type v, index_type case) except? -45.3:

        # self.num_cases checks that the variable is valid

        if case < 0 or case >= self.num_cases(v):
            raise ValueError("case {} is invalid, variable only supports {} "
                             "cases".format(case, self.num_cases(v)))

        return self.cppbqm.linear(self.case_starts_[v] + case)

    def get_quadratic(self, index_type u, index_type v, bint array=False):

        # check that the interaction does in fact exist
        if u < 0 or u >= self.adj_.size():
            raise ValueError("unknown variable")
        if v < 0 or v >= self.adj_.size():
            raise ValueError("unknown variable")

        it = lower_bound(self.adj_[u].begin(), self.adj_[u].end(), v)
        if it == self.adj_[u].end() or deref(it) != v:
            raise ValueError("there is no interaction between given variables")

        cdef index_type ci
        cdef Py_ssize_t case_u, case_v
        cdef bias_type[:, :] quadratic_view

        if array:
            # build a numpy array
            quadratic = np.zeros((self.num_cases(u), self.num_cases(v)),
                                 dtype=self.dtype)
            quadratic_view = quadratic

            for ci in range(self.case_starts_[u], self.case_starts_[u+1]):

                span = self.cppbqm.neighborhood(ci, self.case_starts_[v])

                while (span.first != span.second and deref(span.first).first < self.case_starts_[v+1]):
                    case_u = ci - self.case_starts_[u]
                    case_v = deref(span.first).first - self.case_starts_[v]
                    quadratic_view[case_u, case_v] = deref(span.first).second

                    inc(span.first)

        else:
            # store in a dict
            quadratic = {}

            for ci in range(self.case_starts_[u], self.case_starts_[u+1]):

                span = self.cppbqm.neighborhood(ci, self.case_starts_[v])

                while (span.first != span.second and deref(span.first).first < self.case_starts_[v+1]):
                    case_u = ci - self.case_starts_[u]
                    case_v = deref(span.first).first - self.case_starts_[v]
                    quadratic[case_u, case_v] = deref(span.first).second

                    inc(span.first)

        # todo: support scipy sparse matrices?

        return quadratic

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef bias_type get_quadratic_case(self,
                                  index_type u, index_type case_u,
                                  index_type v, index_type case_v)  except? -45.3:

        if case_u < 0 or case_u >= self.num_cases(u):
            raise ValueError("case {} is invalid, variable only supports {} "
                             "cases".format(case_u, self.num_cases(u)))

        if case_v < 0 or case_v >= self.num_cases(v):
            raise ValueError("case {} is invalid, variable only supports {} "
                             "cases".format(case_v, self.num_cases(v)))


        cdef index_type cu = self.case_starts_[u] + case_u
        cdef index_type cv = self.case_starts_[v] + case_v

        return self.cppbqm.quadratic(cu, cv)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef Py_ssize_t num_cases(self, Py_ssize_t v=-1) except -1:
        """If v is provided, the number of cases associated with v, otherwise
        the total number of cases in the DQM.
        """
        if v < 0:
            return self.cppbqm.num_variables()

        if v >= self.num_variables():
            raise ValueError("unknown variable {}".format(v))

        return self.case_starts_[v+1] - self.case_starts_[v]

    cpdef Py_ssize_t num_case_interactions(self):
        """The total number of case interactions."""
        return self.cppbqm.num_interactions()

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef Py_ssize_t num_variable_interactions(self) except -1:
        """The total number of case interactions."""
        cdef Py_ssize_t num = 0
        cdef Py_ssize_t v
        for v in range(self.num_variables()):
            num += self.adj_[v].size()
        return num // 2

    cpdef Py_ssize_t num_variables(self):
        """The number of discrete variables in the DQM."""
        return self.adj_.size()

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef Py_ssize_t set_linear(self, index_type v, Numeric[:] biases) except -1:

        # self.num_cases checks that the variable is valid

        if biases.shape[0] != self.num_cases(v):
            raise ValueError('Recieved {} bias(es) for a variable of degree {}'
                             ''.format(biases.shape[0], self.num_cases(v)))

        cdef Py_ssize_t c
        for c in range(biases.shape[0]):
            self._set_linear(self.case_starts_[v] + c, biases[c])

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef Py_ssize_t set_linear_case(self, index_type v, index_type case, bias_type b) except -1:
        
        # self.num_cases checks that the variable is valid

        if case < 0:
            raise ValueError("case should be a positive integer")
        if case >= self.num_cases(v):
            raise ValueError("case {} is invalid, variable only supports {} "
                             "cases".format(case, self.num_cases(v)))

        self._set_linear(self.case_starts_[v] + case, b)

    def set_quadratic(self, index_type u, index_type v, biases):

        # check that the interaction does in fact exist
        if u < 0 or u >= self.adj_.size():
            raise ValueError("unknown variable")
        if v < 0 or v >= self.adj_.size():
            raise ValueError("unknown variable")
        if u == v:
            raise ValueError("there cannot be a quadratic interaction between "
                             "two cases in the same variable")

        cdef index_type case_u, case_v
        cdef index_type cu, cv
        cdef bias_type bias

        cdef Py_ssize_t num_cases_u = self.num_cases(u)
        cdef Py_ssize_t num_cases_v = self.num_cases(v)

        cdef bias_type[:, :] biases_view

        if isinstance(biases, abc.Mapping):
            for (case_u, case_v), bias in biases.items():
                if case_u < 0 or case_u >= num_cases_u:
                    raise ValueError("case {} is invalid, variable only supports {} "
                                     "cases".format(case_u, self.num_cases(u)))

                if case_v < 0 or case_v >= num_cases_v:
                    raise ValueError("case {} is invalid, variable only supports {} "
                                     "cases".format(case_v, self.num_cases(v)))

                cu = self.case_starts_[u] + case_u
                cv = self.case_starts_[v] + case_v

                self.cppbqm.set_quadratic(cu, cv, bias)
        else:
            
            biases_view = np.asarray(biases, dtype=self.dtype).reshape(num_cases_u, num_cases_v)

            for case_u in range(biases_view.shape[0]):
                cu = self.case_starts_[u] + case_u
                for case_v in range(biases_view.shape[1]):
                     cv = self.case_starts_[v] + case_v

                     bias = biases_view[case_u, case_v]

                     if bias:
                         self.cppbqm.set_quadratic(cu, cv, bias)

        # track in adjacency
        low = lower_bound(self.adj_[u].begin(), self.adj_[u].end(), v)
        if low == self.adj_[u].end() or deref(low) != v:
            # need to add
            self.adj_[u].insert(low, v)
            self.adj_[v].insert(
                lower_bound(self.adj_[v].begin(), self.adj_[v].end(), u),
                u)

    cpdef Py_ssize_t set_quadratic_case(self,
                                        index_type u, index_type case_u,
                                        index_type v, index_type case_v,
                                        bias_type bias) except -1:

        # self.num_cases checks that the variables are valid

        if case_u < 0 or case_u >= self.num_cases(u):
            raise ValueError("case {} is invalid, variable only supports {} "
                             "cases".format(case_u, self.num_cases(u)))

        if case_v < 0 or case_v >= self.num_cases(v):
            raise ValueError("case {} is invalid, variable only supports {} "
                             "cases".format(case_v, self.num_cases(v)))

        if u == v:
            raise ValueError("there cannot be a quadratic interaction between "
                             "two cases in the same variable")


        cdef index_type cu = self.case_starts_[u] + case_u
        cdef index_type cv = self.case_starts_[v] + case_v

        self.cppbqm.set_quadratic(cu, cv, bias)

        # track in adjacency
        low = lower_bound(self.adj_[u].begin(), self.adj_[u].end(), v)
        if low == self.adj_[u].end() or deref(low) != v:
            # need to add
            self.adj_[u].insert(low, v)
            self.adj_[v].insert(
                lower_bound(self.adj_[v].begin(), self.adj_[v].end(), u),
                u)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef void _into_numpy_vectors(self, Unsigned[:] starts, bias_type[:] ldata,
                                  Unsigned[:] irow, Unsigned[:] icol, bias_type[:] qdata):
        # we don't do array length checking so be careful! This can segfault
        # if the given arrays are incorrectly sized

        cdef Py_ssize_t vi
        for vi in range(self.num_variables()):
            starts[vi] = self.case_starts_[vi]

        cdef Py_ssize_t ci = 0
        cdef Py_ssize_t qi = 0
        for ci in range(self.cppbqm.num_variables()):
            ldata[ci] = self.cppbqm.linear(ci)

            span = self.cppbqm.neighborhood(ci)
            while span.first != span.second and deref(span.first).first < ci:

                irow[qi] = ci
                icol[qi] = deref(span.first).first
                qdata[qi] = deref(span.first).second

                inc(span.first)
                qi += 1
        

    def to_numpy_vectors(self, bint return_offset=False):
        
        cdef Py_ssize_t num_variables = self.num_variables()
        cdef Py_ssize_t num_cases = self.num_cases()
        cdef Py_ssize_t num_interactions = self.cppbqm.num_interactions()

        # use the minimum sizes of the various index types. We combine for
        # variables and cases and exclude int8 to keep the total number of
        # cases down
        if num_cases < 1 << 16:
            index_dtype = np.uint16
        elif num_cases < 1 << 32:
            index_dtype = np.uint32
        else:
            index_dtype = np.uint64

        starts = np.empty(num_variables, dtype=index_dtype)
        ldata = np.empty(num_cases, dtype=self.dtype)
        irow = np.empty(num_interactions, dtype=index_dtype)
        icol = np.empty(num_interactions, dtype=index_dtype)
        qdata = np.empty(num_interactions, dtype=self.dtype)

        if index_dtype == np.uint16:
            self._into_numpy_vectors[np.uint16_t](starts, ldata, irow, icol, qdata)
        elif index_dtype == np.uint32:
            self._into_numpy_vectors[np.uint32_t](starts, ldata, irow, icol, qdata)
        else:
            self._into_numpy_vectors[np.uint64_t](starts, ldata, irow, icol, qdata)

        if return_offset:
            return starts, ldata, (irow, icol, qdata), self.offset

        return starts, ldata, (irow, icol, qdata)



# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
A composite that scales problem variables as directed. If a scaling value is
not specified, calculates it based on quadratic and bias ranges.

"""

try:
    from dwave.preprocessing import ScaleComposite as _ScaleComposite
except ImportError:
    from dimod.reference.composites._preprocessing import NotFound as _ScaleComposite

from dimod.reference.composites._preprocessing import DeprecatedToPreprocessing


__all__ = ['ScaleComposite']


class ScaleComposite(DeprecatedToPreprocessing, _ScaleComposite):
    pass


ScaleComposite.__doc__ = _ScaleComposite.__doc__



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from typing import Callable, Collection, List, Mapping, Optional

import numpy.random

from dimod.binary_quadratic_model import BinaryQuadraticModel
from dimod.decorators import graph_argument
from dimod.typing import GraphLike, Variable
from dimod.vartypes import SPIN

__all__ = ['frustrated_loop']


Predicate = Callable[[List[Variable]], bool]


@graph_argument('graph')
def frustrated_loop(graph: GraphLike,
                    num_cycles: int,
                    R: float = float('inf'),
                    cycle_predicates: Collection[Predicate] = tuple(),
                    max_failed_cycles: int = 100,
                    planted_solution: Optional[Mapping[Variable, int]] = None,
                    seed: Optional[int] = None,
                    ) -> BinaryQuadraticModel:
    """Generate a frustrated-loop problem.

    A generic frustrated-loop (FL) problem is a sum of Hamiltonians, each generated
    from a single "good" loop, as follows:

    1. Generate a loop by random walking on the specified graph, ``graph``.
    2. If the cycle meets provided predicates, continue; if not, return to  step 1.
    3. Choose one edge of the loop to be anti-ferromagnetic (an interaction value
       of `+1` for the edge) and all others ferromagnetic (`-1`).
    4. Add the loop's interactions to the FL problem.
       If at any time the absolute value of an interaction in the FL problem,
       accumulated on an edge over good loops, exceeds a given maximum, ``R``,
       remove that edge from consideration in the loop generation procedure.

    This is a generic generator of FL problems that encompasses both the original
    FL problem definition [#HJARTL]_ and the limited FL problem
    definition [#KLH]_.

    Args:
        graph:
            The graph to build the frustrated loops on. Either an integer, `n`,
            interpreted as a complete graph of size `n`, a nodes/edges pair, a
            list of edges, or a NetworkX graph.

        num_cyles:
            Desired number of frustrated cycles.

        R:
            Maximum absolute interaction weight an edge can accumulate from good
            cycles.

        cycle_predicates:
            An iterable of functions, which should accept a cycle and return a bool.

        max_failed_cycles:
            Maximum number of failures to find a cycle before terminating.

        planted_solution:
            Other solutions with the same energy may exist, but the energy value
            of the (possibly degenerate) ground state will hold. If None,
            planted_solution is: ``{v: -1 for v in graph}``

        seed: Random seed.

    .. [#HJARTL] Hen, I., J. Job, T. Albash, T.F. Rønnow, M. Troyer, D. Lidar. Probing for quantum
        speedup in spin glass problems with planted solutions. https://arxiv.org/abs/1502.01663v2

    .. [#KLH] King, A.D., T. Lanting, R. Harris. Performance of a quantum annealer on range-limited
        constraint satisfaction problems. https://arxiv.org/abs/1502.02098

    """
    nodes, edges = graph
    if num_cycles <= 0:
        raise ValueError("num_cycles should be a positive integer")
    if R <= 0:
        raise ValueError("R should be a positive integer")
    if max_failed_cycles <= 0:
        raise ValueError("max_failed_cycles should be a positive integer")
    if planted_solution is None:
        planted_solution = {v: -1 for v in nodes}
    r = numpy.random.RandomState(seed)

    adj = {v: set() for v in nodes}
    for u, v in edges:
        if u in adj:
            adj[u].add(v)
        else:
            adj[u] = {v}
        if v in adj:
            adj[v].add(u)
        else:
            adj[v] = {u}
    bqm = BinaryQuadraticModel({v: 0.0 for v in nodes}, {edge: 0.0 for edge in edges}, 0.0, SPIN)

    failed_cycles = 0
    good_cycles = 0
    while good_cycles < num_cycles and failed_cycles < max_failed_cycles:

        cycle = _random_cycle(adj, r)

        # if the cycle failed or it is otherwise invalid, mark as failed and continue
        if cycle is None or not all(pred(cycle) for pred in cycle_predicates):
            failed_cycles += 1
            continue
        good_cycles += 1

        # If its a good cycle, modify J with it, according to plated solution
        cycle_J = {}
        for i, c in enumerate(cycle):
            u, v = cycle[i - 1], cycle[i]
            J = -planted_solution[u]*planted_solution[v]
            cycle_J[(u, v)] = J

        # randomly select an edge and flip it
        idx = r.randint(len(cycle))
        cycle_J[(cycle[idx - 1], cycle[idx])] *= -1

        # update the bqm
        bqm.add_interactions_from(cycle_J)
        for u, v in cycle_J:
            if abs(bqm.adj[u][v]) >= R:
                adj[u].remove(v)
                adj[v].remove(u)

    if good_cycles < num_cycles:
        raise RuntimeError(f"number of found cycles, {good_cycles}, is below " 
                           f"requested {num_cycles}")

    return bqm


def _random_cycle(adj, random_state):
    """Find a cycle using a random graph walk."""

    # step through idx values in adj to pick a random one, random.choice does not work on dicts
    n = random_state.randint(len(adj))
    for idx, v in enumerate(adj):
        if idx == n:
            break
    start = v

    walk = [start]
    visited = {start: 0}

    while True:
        if len(walk) > 1:
            # as long as we don't step back one we won't have any repeated edges
            previous = walk[-2]
            neighbors = [u for u in adj[walk[-1]] if u != previous]
        else:
            neighbors = list(adj[walk[-1]])

        if not neighbors:
            # we've walked into a dead end
            return None

        # get a random neighbor
        u = random_state.choice(neighbors)
        if u in visited:
            # if we've seen this neighbour, then we have a cycle starting from it
            return walk[visited[u]:]
        else:
            # add to walk and keep moving
            walk.append(u)
            visited[u] = len(visited)



# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Unified warnings/exceptions for stuff moved to dwave-preprocessing."""

import warnings

__all__ = ['NotFound', 'DeprecatedToPreprocessing']


class NotFound:
    pass


class DeprecatedToPreprocessing:
    def __init__(self, *args, **kwargs):
        if isinstance(self, NotFound):
            # we recommend --no-deps because its dependencies are the same as
            # dimods and it would be a circular install otherwise
            raise TypeError(
                f"{type(self).__name__!r} has been moved to dwave-preprocessing. "
                "You must install dwave-preprocessing in order to use it. "
                "You can do so with "
                "'pip install \"dwave-preprocessing<0.4\" --no-deps'.",
                )

        # otherwise warn about it's new location but let it proceed
        warnings.warn(
            f"{type(self).__name__!s} has been moved to dwave-preprocessing "
            "and will be removed from dimod 0.11.0. To avoid this warning, "
            "import it with 'from dwave.preprocessing import "
            f"{type(self).__name__!s}'.",
            DeprecationWarning, stacklevel=2
            )

        super().__init__(*args, **kwargs)



# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import typing

from pyparsing import *
from collections import defaultdict, namedtuple
from dimod.vartypes import Vartype
from dimod.quadratic import QuadraticModel

MINIMIZE = 1
MAXIMIZE = -1

obj_senses = {"max": MAXIMIZE, "maximum": MAXIMIZE, "maximize": MAXIMIZE,  "min": MINIMIZE, "minimum": MINIMIZE,
              "minimize": MINIMIZE}

GEQ = -1
EQ = 1
LEQ = 1

constraint_senses = {"<": LEQ, "<=": LEQ, "=<": LEQ, "=": EQ,  ">": GEQ, ">=": GEQ, "=>": GEQ}
constraint_symbols = {"<": "<=", "<=": "<=", "=<": "<=", "=": "==",  ">": ">=", ">=": ">=", "=>": ">="}
infinity = 1E30


def make_lp_grammar() -> And:
    """Build the grammar of LP files"""
    # name char ranges for objective, constraint or variable
    all_name_chars = alphanums + "!\"#$%&()/,.;?@_'`{}|~"
    first_char = remove_strings(all_name_chars, nums)
    name = Word(first_char, all_name_chars, max=255)

    # keywords in the LP file
    keywords = ["inf", "infinity",
                "max", "maximum", "maximize",
                "min", "minimum", "minimize",
                "s.t.", "st",
                "bound", "bounds",
                "bin", "binaries", "binary",
                "gen", "general", "generals",
                "end"]

    py_keyword = MatchFirst(map(CaselessKeyword, keywords))
    valid_name = ~py_keyword + name
    valid_name = valid_name.setResultsName("name")

    # second variable for quadratic terms
    second_name = ~py_keyword + name
    second_name = second_name.setResultsName("second_var_name")

    squared_name = ~py_keyword + name
    squared_name = squared_name.setResultsName('squared_name')

    colon = Suppress(oneOf(": ::"))
    plus_minus = oneOf("+ -")
    inf = oneOf("inf infinity", caseless=True)
    number = Word(nums + ".")
    sense = oneOf("< <= =< = > >= =>").setResultsName("sense")

    # section tags
    obj_tag_max = oneOf("max maximum maximize", caseless=True)
    obj_tag_min = oneOf("min minimum minimize", caseless=True)
    obj_tag = (obj_tag_max | obj_tag_min).setResultsName("obj_sense")

    constraints_tag = oneOf(["subj to", "subject to", "s.t.", "st"], caseless=True)

    bounds_tag = oneOf("bound bounds", caseless=True)
    bin_tag = oneOf("bin binaries binary", caseless=True)
    gen_tag = oneOf("gen general generals", caseless=True)
    end_tag = CaselessLiteral("end")

    # coefficient on a variable (includes sign)
    first_var_coef = Optional(plus_minus, "+") + Optional(number, "1")
    first_var_coef.setParseAction(lambda tokens: eval("".join(tokens)))

    coef = plus_minus + Optional(number, "1")
    coef.setParseAction(lambda tokens: eval("".join(tokens)))

    # variable (coefficient and name)
    first_var = Group(first_var_coef.setResultsName("coef") + valid_name)
    var = Group(coef.setResultsName("coef") + valid_name)

    # linear expression
    lin_expr = first_var + ZeroOrMore(var)
    lin_expr = lin_expr.setResultsName("lin_expr")

    # bilinear expression
    quad_vars = Group(first_var_coef.setResultsName("coef") + valid_name + Literal('*').suppress() + second_name)
    # squared expression
    square_vars = Group(
        first_var_coef.setResultsName("coef") + squared_name + Literal('^').suppress() + Literal('2').suppress())

    quadratic_terms = quad_vars | square_vars
    quad_expr = Optional(Literal('+').suppress()) + Literal('[').suppress() + quadratic_terms + \
                ZeroOrMore(quadratic_terms) + Literal(']').suppress()
    quad_expr = quad_expr.setResultsName("quad_expr")

    # for the objective function the standard is having [ quad expression ] / 2
    quad_expr_obj = Optional(Literal('+').suppress()) + Literal('[').suppress() + quadratic_terms + \
                    ZeroOrMore(quadratic_terms) + Literal(']').suppress() + Literal('/').suppress() + \
                    Literal('2').suppress()
    quad_expr_obj = quad_expr_obj.setResultsName("quad_expr")

    # objective
    objective = obj_tag + Optional(valid_name + colon) + Optional(lin_expr) + Optional(quad_expr_obj)
    objective = objective.setResultsName("objective")

    # constraint rhs
    rhs = Optional(plus_minus, "+") + number
    rhs = rhs.setResultsName("rhs")
    rhs.setParseAction(lambda tokens: eval("".join(tokens)))

    # constraints (can be quadratic)
    constraint_word = Group(Optional(valid_name + colon) + Optional(lin_expr) + Optional(quad_expr) + sense + rhs)
    constraints = ZeroOrMore(constraint_word)
    constraints = constraints.setResultsName("constraints")

    # bounds
    signed_inf = (plus_minus + inf).setParseAction(lambda tokens: (tokens[0] == "+") * infinity)
    signed_number = (Optional(plus_minus, "+") + number).setParseAction(lambda tokens: eval("".join(tokens)))
    number_or_inf = (signed_number | signed_inf).setResultsName("numberOrInf")

    # splitting the bounds in left inequality and right inequality
    left_ineq = number_or_inf + sense
    right_ineq = sense + number_or_inf
    bounded_var = Group(Optional(left_ineq).setResultsName("leftbound") + valid_name +
                        Optional(right_ineq).setResultsName("rightbound"))
    free_var = Group(valid_name + Literal("free"))

    bounds_word = free_var | bounded_var
    bounds = bounds_tag + ZeroOrMore(bounds_word).setResultsName("bounds")

    # generals (integer variables)
    generals = gen_tag + ZeroOrMore(valid_name).setResultsName("generals")

    # binaries (binary variables)
    binaries = bin_tag + ZeroOrMore(valid_name).setResultsName("binaries")

    var_info = ZeroOrMore(bounds | generals | binaries)

    # full LP file grammar
    grammar = Optional(objective) + constraints_tag + constraints + var_info + end_tag

    # commenting
    comment_style = Literal("\\") + restOfLine
    grammar.ignore(comment_style)

    return grammar


def remove_strings(string, strings_to_remove):
    """Replace an iterable of strings in removables
       if removables is a string, each character is removed """
    for r in strings_to_remove:
        try:
            string = string.replace(r, "")
        except TypeError:
            raise TypeError("Strings_to_remove contains a non-string element")
    return string


def get_variables_from_parsed_lp(parse_output: ParseResults,
                                 lower_bound_default: typing.Optional[int] = None,
                                 upper_bound_default: typing.Optional[int] = None) -> QuadraticModel:
    """Return a quadratic model containing all the variables included in the CQM

    Args:
        parse_output: the parse results encoding the LP file
        lower_bound_default: the lower bound of the integer variables, in case they are not specified
        upper_bound_default: the upper bound of the integer variables in case they are not specified

    Returns:
        the quadratic model with all variables
    """

    obj = QuadraticModel()

    all_vars = set()
    # default variable type for LP file
    # should be continuous, even though
    # they are not supported in CQMs
    Var = namedtuple('Variable', ['vartype', 'lb', 'ub'])
    variables_info = defaultdict(lambda: Var(vartype="c", lb=lower_bound_default, ub=upper_bound_default))

    # scan the objective
    for oe in parse_output.objective:
        if isinstance(oe, str):
            continue

        else:
            if len(oe) == 2:
                if oe.name != "":
                    all_vars.add(oe.name[0])
                else:
                    all_vars.add(oe.squared_name[0])
            elif len(oe) == 3:
                all_vars.update([oe.name[0], oe.second_var_name[0]])

    # scan the constraints
    for c in parse_output.constraints:

        # scan linear terms
        if c.lin_expr:
            all_vars.update([le.name[0] for le in c.lin_expr])

        # scan quadratic terms of constraints
        if c.quad_expr:
            for qe in c.quad_expr:
                if len(qe) == 3:
                    all_vars.update([qe.name[0] for qe in c.quad_expr])
                    all_vars.update([qe.second_var_name[0] for qe in c.quad_expr])
                elif len(qe) == 2:
                    all_vars.add(qe.squared_name[0])

    # scan the bounds
    for b in parse_output.bounds:

        n = b.name[0]

        # if b.free, default is fine
        if b.leftbound:
            if constraint_senses[b.sense] <= 0:  # NUM >= var
                variables_info[n] = variables_info[n]._replace(ub=b.leftbound[0])
            if constraint_senses[b.sense] >= 0:  # NUM <= var
                variables_info[n] = variables_info[n]._replace(lb=b.leftbound[0])

        if b.rightbound:
            if constraint_senses[b.sense] <= 0:  # var >= NUM
                variables_info[n] = variables_info[n]._replace(lb=b.rightbound[1])

            if constraint_senses[b.sense] >= 0:  # var <= NUM
                variables_info[n] = variables_info[n]._replace(ub=b.rightbound[1])

    # check the binary variables:
    for n in parse_output.binaries:
        variables_info[n] = variables_info[n]._replace(vartype=Vartype.BINARY)

    # check for integer variables
    for n in parse_output.generals:
        variables_info[n] = variables_info[n]._replace(vartype=Vartype.INTEGER)

    for n, var_info in variables_info.items():
        if var_info.vartype is Vartype.BINARY:
            obj.add_variable(Vartype.BINARY, n)
        elif var_info.vartype is Vartype.INTEGER:
            lb = var_info.lb
            ub = var_info.ub

            if lb is not None:
                obj.add_variable(Vartype.INTEGER, n, lower_bound=lb, upper_bound=ub)
            else:
                obj.add_variable(Vartype.INTEGER, n, upper_bound=ub)
        else:
            raise ValueError("Unexpected Vartype: {} for variable: {}".format(var_info.vartype, n))

    return obj



# Copyright 2020 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import collections.abc
import functools
import warnings

import dimod

__all__ = ['load_sampler_bqm_tests']

# in the future we may want a way for folks to load new BQM classes into this
# framework
BQM_SUBCLASSES = [dimod.Float32BQM,
                  dimod.Float64BQM,
                  dimod.DictBQM,
                  dimod.AdjVectorBQM,
                  dimod.AdjDictBQM,
                  ]


def test_sample(self, sampler, bqm):
    sampleset = sampler.sample(bqm)

    self.assertEqual(set(sampleset.variables), set(bqm.variables))
    self.assertIs(sampleset.vartype, bqm.vartype)
    dimod.testing.assert_sampleset_energies(sampleset, bqm, precision=4)


def test_sample_ising(self, sampler, h, J):
    sampleset = sampler.sample_ising(h, J)

    self.assertEqual(set(sampleset.variables), set(h).union(*J))
    self.assertIs(sampleset.vartype, dimod.SPIN)

    for sample, en in sampleset.data(['sample', 'energy']):
        self.assertAlmostEqual(dimod.ising_energy(sample, h, J), en)


def test_sample_qubo(self, sampler, Q):
    sampleset = sampler.sample_qubo(Q)

    self.assertEqual(set(sampleset.variables), set().union(*Q))
    self.assertIs(sampleset.vartype, dimod.BINARY)

    for sample, en in sampleset.data(['sample', 'energy']):
        self.assertAlmostEqual(dimod.qubo_energy(sample, Q), en)


class TestCaseOverloadWarning(UserWarning):
    pass


# this is functionally quite similar to functools.partialmethod but it avoids
# the confusing error message on a failed test that partial creates.
def parameterized_method(f, *args, specifiers=None):
    @functools.wraps(f)
    def method(self):
        return f(self, *args)

    if specifiers:
        method.__name__ += '_' + '_'.join(specifiers)

    return method


def create_bqm_tests(sampler, max_num_variables=None):
    """Yields test methods that can be added to a TestCase."""

    if max_num_variables is None:
        max_num_variables = float('inf')
    elif max_num_variables < 0:
        raise ValueError('max_num_variables should be a positive int')

    # empty

    specifiers = ['empty', sampler.__name__]

    h = {}
    J = {}

    yield parameterized_method(test_sample_ising, sampler(), h, J,
                               specifiers=specifiers)

    Q = {}

    yield parameterized_method(test_sample_qubo, sampler(), Q,
                               specifiers=specifiers)

    for BQM in BQM_SUBCLASSES:
        yield parameterized_method(test_sample, sampler(), BQM(h, J, 1.5, 'SPIN'),
                                   specifiers=['spin', BQM.__name__] + specifiers)
        yield parameterized_method(test_sample, sampler(), BQM.from_qubo(Q, 1.5),
                                   specifiers=['binary', BQM.__name__] + specifiers)

    # 1 variable

    if max_num_variables < 1:
        return

    u = (('a',),)  # complicated label

    specifiers = ['1var', sampler.__name__]

    h = {u: 6.0}
    J = {}

    yield parameterized_method(test_sample_ising, sampler(), h, J,
                               specifiers=specifiers)

    Q = {(u, u): 6.0}

    yield parameterized_method(test_sample_qubo, sampler(), Q,
                               specifiers=specifiers)

    for BQM in BQM_SUBCLASSES:
        yield parameterized_method(test_sample, sampler(), BQM(h, J, 1.5, 'SPIN'),
                                   specifiers=['spin', BQM.__name__] + specifiers)
        yield parameterized_method(test_sample, sampler(), BQM.from_qubo(Q, 1.5),
                                   specifiers=['binary', BQM.__name__] + specifiers)

    if max_num_variables < 2:
        return

    v = 0  # new variable

    specifiers = ['1path', sampler.__name__]

    h = {u: 6.0}
    J = {(u, v): -3}

    yield parameterized_method(test_sample_ising, sampler(), h, J,
                               specifiers=specifiers)

    Q = {(u, u): 6.0, (u, v): -3}

    yield parameterized_method(test_sample_qubo, sampler(), Q,
                               specifiers=specifiers)

    for BQM in BQM_SUBCLASSES:
        yield parameterized_method(test_sample, sampler(), BQM(h, J, 3, 'SPIN'),
                                   specifiers=['spin', BQM.__name__] + specifiers)
        yield parameterized_method(test_sample, sampler(), BQM.from_qubo(Q, 8),
                                   specifiers=['binary', BQM.__name__] + specifiers)

    if max_num_variables < 3:
        return

    w = 'c'  # new variable

    specifiers = ['2path', sampler.__name__]

    h = {u: 6.0}
    J = {(u, v): -3, (v, w): 105}

    yield parameterized_method(test_sample_ising, sampler(), h, J,
                               specifiers=specifiers)

    Q = {(u, u): 6.0, (u, v): -3, (v, w): 105}

    yield parameterized_method(test_sample_qubo, sampler(), Q,
                               specifiers=specifiers)

    for BQM in BQM_SUBCLASSES:
        yield parameterized_method(test_sample, sampler(), BQM(h, J, -4., 'SPIN'),
                                   specifiers=['spin', BQM.__name__] + specifiers)
        yield parameterized_method(test_sample, sampler(), BQM.from_qubo(Q, 16),
                                   specifiers=['binary', BQM.__name__] + specifiers)


def load_sampler_bqm_tests(sampler, max_num_variables=None,
                           suppress_overload_warning=False):
    """Populate the decorated TestCase with sampler tests using small BQMs.

    Args:
        sampler (obj/callable):
            A :class:`.Sampler`. If given as an object, the same sampler will
            be used for all tests, if given as a callable a new sampler will
            be instantiated for each test.

        max_num_variables (int, optional):
            The maximum BQM size to be tested. Defaults to unlimited.

        suppress_overload_warning (bool, optional, default=False):
            If True and the decorated class has a method matching one of the
            ones generated by `load_sampler_bqm_tests`, a warning will be
            emitted.

    Example:

        For example, if one wanted to test the :class:`.ExactSolver`,

        .. code-block:: python

            import unittest

            import dimod.testing

            @dimod.testing.load_sampler_bqm_tests(dimod.ExactSolver)
            class TestExactSolver(unittest.TestCase):
                pass

            unittest.main()

    """

    if isinstance(sampler, collections.abc.Callable):
        sampler_factory = sampler
    else:
        def sampler_factory():
            return sampler
        sampler_factory.__name__ = str(sampler)

    def decorator(cls):
        for test_method in create_bqm_tests(sampler_factory,
                                            max_num_variables=max_num_variables):

            if not hasattr(cls, test_method.__name__):
                setattr(cls, test_method.__name__, test_method)
            elif not suppress_overload_warning:
                msg = ('{} has been overloaded. If this is the desired '
                       'behaviour, this warning can be suppressed with the '
                       'suppress_overload_warning kwarg.')
                warnings.warn(msg, TestCaseOverloadWarning)

        return cls

    return decorator



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
A class and utilities for encoding variable objects.

The :class:`Variables` class is intended to be used as an attribute of other
classes, such as :class:`.DiscreteQuadraticModel` and :class:`.SampleSet`.

The requirements for the class are:
    *   Have a minimal memory footprint when the variables are labeled `[0, n)`
    *   Behave like a list for iteration and getting items
    *   Behave like a set for determining if it contains an item
    *   Constant time for finding the index of an item

"""

import collections.abc as abc
import io
import warnings

from numbers import Integral, Number
from operator import eq
from pprint import PrettyPrinter

from dimod.cyvariables import cyVariables


__all__ = ['Variables']


def serialize_variable(v):
    # want to handle things like numpy numbers and fractions that do not
    # serialize so easy
    if isinstance(v, Integral):
        return int(v)
    elif isinstance(v, Number):
        return float(v)
    elif isinstance(v, str):
        return v
    elif isinstance(v, abc.Collection):
        return tuple(iter_serialize_variables(v))
    else:
        return v


def iter_serialize_variables(variables):
    yield from map(serialize_variable, variables)


def deserialize_variable(v):
    if isinstance(v, abc.Collection) and not isinstance(v, str):
        return tuple(iter_deserialize_variables(v))
    else:
        return v


def iter_deserialize_variables(variables):
    yield from map(deserialize_variable, variables)


class Variables(cyVariables, abc.Set, abc.Sequence):
    """Set-like and list-like variables tracking.

    Args:
        iterable (iterable):
            An iterable of labels. Duplicate labels are ignored. All labels
            must be hashable.

    """
    def __eq__(self, other):
        if isinstance(other, abc.Sequence):
            return len(self) == len(other) and all(map(eq, self, other))
        elif isinstance(other, abc.Set):
            return not (self ^ other)
        else:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        stream = io.StringIO()
        stream.write(type(self).__name__)
        stream.write('(')
        if self:
            if self.is_range and len(self) > 10:
                # 10 is arbitrary, but the idea is we want to truncate
                # longer variables that are integer-labelled
                stream.write(repr(range(len(self))))
            else:
                stream.write('[')
                iterator = iter(self)
                stream.write(repr(next(iterator)))
                for v in iterator:
                    stream.write(', ')
                    stream.write(repr(v))
                stream.write(']')
        stream.write(')')
        return stream.getvalue()

    @property
    def is_range(self):
        """Return True if the variables are labeled `[0,n)`."""
        return self._is_range()

    def to_serializable(self):
        """Return an object that is json-serializable.

        Returns:
            list: A list of json-serializable objects. Handles some
            common cases like NumPy scalars.
            See :func:`iter_serialize_variables`.

        """
        return list(iter_serialize_variables(self))


# register the various objects with prettyprint
def _pprint_variables(printer, variables, stream, indent, *args, **kwargs):
    if not variables or variables.is_range:
        stream.write(repr(variables))
    else:
        indent += stream.write(type(variables).__name__)
        indent += stream.write('(')
        printer._pprint_list(variables, stream, indent, *args, **kwargs)
        indent += stream.write(')')


try:
    PrettyPrinter._dispatch[Variables.__repr__] = _pprint_variables
except AttributeError:
    # we're using some internal stuff in PrettyPrinter so let's silently fail
    # for that
    pass



# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import operator

from copy import deepcopy

cimport cython

from cython.operator cimport preincrement as inc, dereference as deref
from libc.math cimport ceil, floor

from dimod.binary.cybqm cimport cyBQM
from dimod.cyutilities cimport as_numpy_float, ConstInteger, ConstNumeric, cppvartype
from dimod.libcpp cimport cppvartype_info
from dimod.sampleset import as_samples
from dimod.variables import Variables
from dimod.vartypes import as_vartype, Vartype


cdef class cyQM_template(cyQMBase):
    def __init__(self):
        self.dtype = np.dtype(BIAS_DTYPE)
        self.index_dtype = np.dtype(INDEX_DTYPE)
        self.variables = Variables()

    def __deepcopy__(self, memo):
        cdef cyQM_template new = type(self)()
        new.cppqm = self.cppqm
        new.variables = deepcopy(self.variables, memo)
        memo[id(self)] = new
        return new

    @property
    def offset(self):
        """Constant energy offset associated with the model."""
        return as_numpy_float(self.cppqm.offset())

    @offset.setter
    def offset(self, bias_type offset):
        cdef bias_type *b = &(self.cppqm.offset())
        b[0] = offset

    cdef void _add_linear(self, Py_ssize_t vi, bias_type bias):
        # unsafe version of .add_linear
        cdef bias_type *b = &(self.cppqm.linear(vi))
        b[0] += bias

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _ilinear(self):
        """Return the linear biases in a numpy array."""
        cdef bias_type[:] ldata = np.empty(self.num_variables(), dtype=self.dtype)
        cdef Py_ssize_t vi
        for vi in range(self.num_variables()):
            ldata[vi] = self.cppqm.linear(vi)
        return ldata

    def _ilower_triangle(self, Py_ssize_t vi):
        cdef Py_ssize_t degree = self.cppqm.num_interactions(vi)

        dtype = np.dtype([('v', self.index_dtype), ('b', self.dtype)],
                         align=False)
        neighbors = np.empty(degree, dtype=dtype)

        cdef index_type[:] index_view = neighbors['v']
        cdef bias_type[:] bias_view = neighbors['b']

        span = self.cppqm.neighborhood(vi)
        cdef Py_ssize_t i = 0
        while span.first != span.second:
            if deref(span.first).first > vi:
                break
            index_view[i] = deref(span.first).first
            bias_view[i] = deref(span.first).second

            i += 1
            inc(span.first)

        return neighbors[:i]

    def _ilower_triangle_load(self, Py_ssize_t vi, Py_ssize_t num_neighbors, buff):
        dtype = np.dtype([('v', self.index_dtype), ('b', self.dtype)],
                         align=False)

        arr = np.frombuffer(buff[:dtype.itemsize*num_neighbors], dtype=dtype)
        cdef const index_type[:] index_view = arr['v']
        cdef const bias_type[:] bias_view = arr['b']

        cdef Py_ssize_t i
        for i in range(num_neighbors):
            self.cppqm.add_quadratic(vi, index_view[i], bias_view[i])

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _ivartypes(self):
        cdef Py_ssize_t num_variables = self.num_variables()

        # we could use the bias_type size to determine the vartype dtype to get
        # more alignment, but it complicates the code, so let's keep it simple
        dtype = np.dtype([('vartype', np.int8),
                          ('lb', BIAS_DTYPE), ('ub', BIAS_DTYPE)],
                         align=False)
        arr = np.empty(num_variables, dtype)

        cdef np.int8_t[:] vartype_view = arr['vartype']
        cdef bias_type[:] lb_view = arr['lb']
        cdef bias_type[:] ub_view = arr['ub']

        cdef Py_ssize_t vi
        for vi in range(self.num_variables()):
            vartype_view[vi] = self.cppqm.vartype(vi)
            lb_view[vi] = self.cppqm.lower_bound(vi)
            ub_view[vi] = self.cppqm.upper_bound(vi)

        return memoryview(arr).cast('B')

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _ivartypes_load(self, buff, Py_ssize_t num_variables):
        if self.num_variables():
            raise RuntimeError("cannot load vartypes into a model with variables")

        # use the bias size to determine the vartype size since we're not
        # letting numpy handle the alignment
        dtype = np.dtype([('vartype', np.int8),
                          ('lb', BIAS_DTYPE), ('ub', BIAS_DTYPE)],
                         align=False)

        arr = np.frombuffer(buff[:dtype.itemsize*num_variables], dtype=dtype)
        cdef const np.int8_t[:] vartype_view = arr['vartype']
        cdef const bias_type[:] lb_view = arr['lb']
        cdef const bias_type[:] ub_view = arr['ub']

        cdef Py_ssize_t vi
        cdef cppVartype cpp_vartype
        for vi in range(num_variables):
            # I hate that we have to do this manually, but cython doesn't
            # really like static_cast in this context
            if vartype_view[vi] == 0:
                cpp_vartype = cppVartype.BINARY
            elif vartype_view[vi] == 1:
                cpp_vartype = cppVartype.SPIN
            elif vartype_view[vi] == 2:
                cpp_vartype = cppVartype.INTEGER
            else:
                raise RuntimeError
            self.cppqm.add_variable(cpp_vartype, lb_view[vi], ub_view[vi])

        while self.variables.size() < self.cppqm.num_variables():
            self.variables._append()

    cdef void _set_linear(self, Py_ssize_t vi, bias_type bias):
        # unsafe version of .set_linear
        cdef bias_type *b = &(self.cppqm.linear(vi))
        b[0] = bias

    def add_linear(self, v, bias_type bias):
        cdef Py_ssize_t vi = self.variables.index(v)
        self._add_linear(vi, bias)

    def add_linear_from_array(self, ConstNumeric[:] linear):
        cdef Py_ssize_t length = linear.shape[0]
        cdef Py_ssize_t vi

        if self.variables._is_range():
            if length > self.num_variables():
                raise ValueError("variables must already exist")
            for vi in range(length):
                self._add_linear(vi, linear[vi])
        else:
            for vi in range(length):
                self.add_linear(vi, linear[vi])

    def add_quadratic(self, u, v, bias_type bias):
        cdef Py_ssize_t ui = self.variables.index(u)
        cdef Py_ssize_t vi = self.variables.index(v)

        if ui == vi and (self.cppqm.vartype(ui) == cppVartype.SPIN
                         or self.cppqm.vartype(ui) == cppVartype.BINARY):
            raise ValueError(f"{u!r} cannot have an interaction with itself")

        self.cppqm.add_quadratic(ui, vi, bias)

    def add_quadratic_from_arrays(self,
                                  ConstInteger[::1] irow, ConstInteger[::1] icol,
                                  ConstNumeric[::1] qdata):
        if not irow.shape[0] == icol.shape[0] == qdata.shape[0]:
            raise ValueError("quadratic vectors should be equal length")
        cdef Py_ssize_t length = irow.shape[0]

        cdef Py_ssize_t vi
        if self.variables._is_range():
            if length > self.num_variables():
                raise ValueError("variables must already exist")

            for vi in range(length):
                self.cppqm.add_quadratic(irow[vi], icol[vi], qdata[vi])
        else:
            for vi in range(length):
                self.add_quadratic(irow[vi], icol[vi], qdata[vi])

    def add_quadratic_from_iterable(self, quadratic):
        cdef Py_ssize_t ui, vi
        cdef bias_type bias

        for u, v, bias in quadratic:
            ui = self.variables.index(u)
            vi = self.variables.index(v)

            if ui == vi and (self.cppqm.vartype(ui) == cppVartype.SPIN
                             or self.cppqm.vartype(ui) == cppVartype.BINARY):
                raise ValueError(f"{u!r} cannot have an interaction with itself")
            
            self.cppqm.add_quadratic(ui, vi, bias)

    def add_variable(self, vartype, label=None, *, lower_bound=None, upper_bound=None):
        # as_vartype will raise for unsupported vartypes
        vartype = as_vartype(vartype, extended=True)
        cdef cppVartype cppvartype = self.cppvartype(vartype)

        cdef bias_type lb
        cdef bias_type ub

        cdef Py_ssize_t vi
        if label is not None and self.variables.count(label):
            # it already exists, so check that vartype matches
            vi = self.variables.index(label)
            if self.cppqm.vartype(vi) != cppvartype:
                raise TypeError(f"variable {label!r} already exists with a different vartype")
            if cppvartype != cppVartype.BINARY and cppvartype != cppVartype.SPIN:
                if lower_bound is not None:
                    lb = lower_bound
                    if lb != self.cppqm.lower_bound(vi):
                        raise ValueError(
                            f"the specified lower bound, {lower_bound}, for "
                            f"variable {label!r} is different than the existing lower "
                            f"bound, {int(self.cppqm.lower_bound(vi))}")
                if upper_bound is not None:
                    ub = upper_bound
                    if ub != self.cppqm.upper_bound(vi):
                        raise ValueError(
                            f"the specified upper bound, {upper_bound}, for "
                            f"variable {label!r} is different than the existing upper "
                            f"bound, {int(self.cppqm.upper_bound(vi))}")

            return label

        if cppvartype == cppVartype.BINARY or cppvartype == cppVartype.SPIN:
            # in this case we just ignore the provided values
            lb = cppvartype_info[bias_type].default_min(cppvartype)
            ub = cppvartype_info[bias_type].default_max(cppvartype)
        elif cppvartype == cppVartype.INTEGER:
            if lower_bound is None:
                lb = cppvartype_info[bias_type].default_min(cppvartype)
            else:
                lb = lower_bound
                if lb < cppvartype_info[bias_type].min(cppvartype):
                    raise ValueError(f"lower_bound cannot be less than {cppvartype_info[bias_type].min(cppvartype)}")

            if upper_bound is None:
                ub = cppvartype_info[bias_type].default_max(cppvartype)
            else:
                ub = upper_bound
                if ub > cppvartype_info[bias_type].max(cppvartype):
                    raise ValueError(f"upper_bound cannot be greater than {cppvartype_info[bias_type].max(cppvartype)}")
            
            if lb > ub:
                raise ValueError("lower_bound must be less than or equal to upper_bound")

            if ceil(lb) > floor(ub):
                raise ValueError("there must be at least one valid integer between lower_bound and upper_bound")
        else:
            raise RuntimeError("unknown vartype")

        self.cppqm.add_variable(cppvartype, lb, ub)

        self.variables._append(label)

        assert self.cppqm.num_variables() == self.variables.size()

        return self.variables.at(-1)

    def change_vartype(self, vartype, v):
        vartype = as_vartype(vartype, extended=True)
        cdef Py_ssize_t vi = self.variables.index(v)
        try:
            self.cppqm.change_vartype(self.cppvartype(vartype), vi)
        except RuntimeError:
            # c++ logic_error
            raise TypeError(f"cannot change vartype {self.vartype(v).name!r} "
                            f"to {vartype.name!r}") from None

    cdef cppVartype cppvartype(self, object vartype) except? cppVartype.SPIN:
        return cppvartype(vartype)

    cdef const cppQuadraticModel[bias_type, index_type]* data(self):
        """Return a pointer to the C++ QuadraticModel."""
        return &self.cppqm

    def degree(self, v):
        cdef Py_ssize_t vi = self.variables.index(v)
        return self.cppqm.num_interactions(vi)

    cdef np.float64_t[::1] _energies(self, ConstNumeric[:, ::1] samples, object labels):
        cdef Py_ssize_t num_samples = samples.shape[0]
        cdef Py_ssize_t num_variables = samples.shape[1]

        if num_variables != len(labels):
            # as_samples should never return inconsistent sizes, but we do this
            # check because the boundscheck is off and we otherwise might get
            # segfaults
            raise RuntimeError("as_samples returned an inconsistent samples/variables")

        # get the indices of the QM variables. Use -1 to signal that a
        # variable's index has not yet been set
        cdef Py_ssize_t[::1] qm_to_sample = np.full(self.num_variables(), -1, dtype=np.intp)
        cdef Py_ssize_t si
        for si in range(num_variables):
            v = labels[si]
            if self.variables.count(v):
                qm_to_sample[self.variables.index(v)] = si

        # make sure that all of the QM variables are accounted for
        for si in range(self.num_variables()):
            if qm_to_sample[si] == -1:
                raise ValueError(f"missing variable {self.variables[si]!r} in sample(s)")

        cdef np.float64_t[::1] energies = np.empty(num_samples, dtype=np.float64)

        # alright, now let's calculate some energies!
        cdef Py_ssize_t ui, vi
        for si in range(num_samples):
            # offset
            energies[si] = self.cppqm.offset()

            for ui in range(self.num_variables()):
                # linear
                energies[si] += self.cppqm.linear(ui) * samples[si, qm_to_sample[ui]];

                span = self.cppqm.neighborhood(ui)
                while span.first != span.second and deref(span.first).first <= ui:
                    vi = deref(span.first).first

                    energies[si] += deref(span.first).second * samples[si, qm_to_sample[ui]] * samples[si, qm_to_sample[vi]]

                    inc(span.first)

        return energies

    def energies(self, samples_like, dtype=None):
        samples, labels = as_samples(samples_like)

        # we need contiguous and unsigned. as_samples actually enforces contiguous
        # but no harm in double checking for some future-proofness
        samples = np.ascontiguousarray(
                samples,
                dtype=f'i{samples.dtype.itemsize}' if np.issubdtype(samples.dtype, np.unsignedinteger) else None,
                )

        # Cython really should be able to figure the type out, but for some reason
        # it fails, so we just dispatch manually
        if samples.dtype == np.float64:
            return np.asarray(self._energies[np.float64_t](samples, labels), dtype=dtype)
        elif samples.dtype == np.float32:
            return np.asarray(self._energies[np.float32_t](samples, labels), dtype=dtype)
        elif samples.dtype == np.int8:
            return np.asarray(self._energies[np.int8_t](samples, labels), dtype=dtype)
        elif samples.dtype == np.int16:
            return np.asarray(self._energies[np.int16_t](samples, labels), dtype=dtype)
        elif samples.dtype == np.int32:
            return np.asarray(self._energies[np.int32_t](samples, labels), dtype=dtype)
        elif samples.dtype == np.int64:
            return np.asarray(self._energies[np.int64_t](samples, labels), dtype=dtype)
        else:
            raise ValueError("unsupported sample dtype")

    @classmethod
    def from_cybqm(cls, cyBQM bqm):

        cdef cyQM_template qm = cls()

        qm.offset = bqm.offset

        # linear
        cdef Py_ssize_t vi
        cdef cppVartype vartype = bqm.cppbqm.vartype()
        for vi in range(bqm.num_variables()):
            qm.cppqm.add_variable(vartype)
            qm._set_linear(vi, bqm.cppbqm.linear(vi))
        qm.variables._extend(bqm.variables)

        # quadratic
        it = bqm.cppbqm.cbegin_quadratic()
        while it != bqm.cppbqm.cend_quadratic():
            qm.cppqm.set_quadratic(deref(it).u, deref(it).v, deref(it).bias)
            inc(it)

        return qm

    def get_linear(self, v):
        cdef Py_ssize_t vi = self.variables.index(v)
        cdef bias_type bias = self.cppqm.linear(vi)
        return as_numpy_float(bias)

    def get_quadratic(self, u, v, default=None):
        cdef Py_ssize_t ui = self.variables.index(u)
        cdef Py_ssize_t vi = self.variables.index(v)

        if ui == vi and (self.cppqm.vartype(ui) == cppVartype.SPIN
                         or self.cppqm.vartype(ui) == cppVartype.BINARY):
            raise ValueError(f"{u!r} cannot have an interaction with itself")

        cdef bias_type bias
        try:
            bias = self.cppqm.quadratic_at(ui, vi)
        except IndexError:
            if default is None:
                # out of range error is automatically converted to IndexError
                raise ValueError(f"{u!r} and {v!r} have no interaction") from None
            bias = default
        return as_numpy_float(bias)

    cpdef bint is_linear(self):
        return self.cppqm.is_linear()

    def iter_neighborhood(self, v):
        cdef Py_ssize_t vi = self.variables.index(v)

        cdef Py_ssize_t ui
        cdef bias_type bias

        span = self.cppqm.neighborhood(vi)
        while span.first != span.second:
            ui = deref(span.first).first
            bias = deref(span.first).second

            yield self.variables.at(ui), as_numpy_float(bias)

            inc(span.first)

    def iter_quadratic(self):
        it = self.cppqm.cbegin_quadratic()
        while it != self.cppqm.cend_quadratic():
            u = self.variables.at(deref(it).u)
            v = self.variables.at(deref(it).v)
            yield u, v, as_numpy_float(deref(it).bias)
            inc(it)

    def lower_bound(self, v):
        cdef Py_ssize_t vi = self.variables.index(v)
        return as_numpy_float(self.cppqm.lower_bound(vi))

    cpdef Py_ssize_t nbytes(self, bint capacity = False):
        return self.cppqm.nbytes(capacity)

    cpdef Py_ssize_t num_interactions(self):
        return self.cppqm.num_interactions()

    cpdef Py_ssize_t num_variables(self):
        return self.cppqm.num_variables()

    def relabel_variables(self, mapping):
        self.variables._relabel(mapping)

    def remove_interaction(self, u, v):
        cdef Py_ssize_t ui = self.variables.index(u)
        cdef Py_ssize_t vi = self.variables.index(v)
        
        if not self.cppqm.remove_interaction(ui, vi):
            raise ValueError(f"{u!r} and {v!r} have no interaction")

    def remove_variable(self, v=None):
        if v is None:
            try:
                v = self.variables[-1]
            except KeyError:
                raise ValueError("cannot pop from an empty model")

        cdef Py_ssize_t vi = self.variables.index(v)
        cdef Py_ssize_t lasti = self.num_variables() - 1

        if vi != lasti:
            # we're removing a variable in the middle of the
            # underlying adjacency. We do this by "swapping" the last variable
            # and v, then popping v from the end
            self.cppqm.swap_variables(vi, lasti)

            # now swap the variable labels
            last = self.variables.at(lasti)
            self.variables._relabel({v: last, last: v})

        # remove last from the cppqm and variables
        self.cppqm.resize(lasti)
        tmp = self.variables._pop()

        assert tmp == v, f"{tmp} == {v}"

        return v

    def relabel_variables_as_integers(self):
        return self.variables._relabel_as_integers()

    def reduce_linear(self, function, initializer=None):
        if self.num_variables() == 0 and initializer is None:
            # feels like this should be a ValueError but python raises
            # TypeError so...
            raise TypeError("reduce_linear() on an empty BQM")

        cdef Py_ssize_t start, vi
        cdef bias_type value, tmp

        if initializer is None:
            start = 1
            value = self.cppqm.linear(0)
        else:
            start = 0
            value = initializer

        # speed up a few common cases
        if function is operator.add:
            for vi in range(start, self.num_variables()):
                value += self.cppqm.linear(vi)
        elif function is max:
            for vi in range(start, self.num_variables()):
                tmp = self.cppqm.linear(vi)
                if tmp > value:
                    value = tmp
        elif function is min:
            for vi in range(start, self.num_variables()):
                tmp = self.cppqm.linear(vi)
                if tmp < value:
                    value = tmp
        else:
            for vi in range(start, self.num_variables()):
                value = function(value, self.cppqm.linear(vi))

        return as_numpy_float(value)

    def reduce_neighborhood(self, u, function, initializer=None):
        cdef Py_ssize_t ui = self.variables.index(u)

        if self.cppqm.num_interactions(ui) == 0 and initializer is None:
            # feels like this should be a ValueError but python raises
            # TypeError so...
            raise TypeError("reduce_neighborhood() on an empty neighbhorhood")

        cdef bias_type value, tmp

        span = self.cppqm.neighborhood(ui)

        if initializer is None:
            value = deref(span.first).second
            inc(span.first)
        else:
            value = initializer

        # speed up a few common cases
        if function is operator.add:
            while span.first != span.second:
                value += deref(span.first).second
                inc(span.first)
        elif function is max:
            while span.first != span.second:
                tmp = deref(span.first).second
                if tmp > value:
                    value = tmp
                inc(span.first)
        elif function is min:
            while span.first != span.second:
                tmp = deref(span.first).second
                if tmp < value:
                    value = tmp
                inc(span.first)
        else:
            while span.first != span.second:
                value = function(value, deref(span.first).second)
                inc(span.first)

        return as_numpy_float(value)

    def reduce_quadratic(self, function, initializer=None):

        if self.cppqm.is_linear() and initializer is None:
            # feels like this should be a ValueError but python raises
            # TypeError so...
            raise TypeError("reduce_quadratic() on a linear model")

        cdef bias_type value, tmp

        start = self.cppqm.cbegin_quadratic()

        if initializer is None:
            value = deref(start).bias
            inc(start)
        else:
            value = initializer

        # handle a few common cases
        if function is operator.add:
            while start != self.cppqm.cend_quadratic():
                value += deref(start).bias
                inc(start)
        elif function is max:
            while start != self.cppqm.cend_quadratic():
                tmp = deref(start).bias
                if tmp > value:
                    value = tmp
                inc(start)
        elif function is min:
            while start != self.cppqm.cend_quadratic():
                tmp = deref(start).bias
                if tmp < value:
                    value = tmp
                inc(start)
        else:
            while start != self.cppqm.cend_quadratic():
                value = function(value, deref(start).bias)
                inc(start)

        return as_numpy_float(value)

    cpdef void scale(self, bias_type scalar):
        self.cppqm.scale(scalar)

    def set_linear(self, v, bias_type bias):
        cdef Py_ssize_t vi = self.variables.index(v)
        self._set_linear(vi, bias)

    def set_lower_bound(self, v, bias_type lb):
        cdef Py_ssize_t vi = self.variables.index(v)
        cdef cppVartype cppvartype = self.cppqm.vartype(vi)

        if cppvartype == cppVartype.BINARY or cppvartype == cppVartype.SPIN:
            raise ValueError(
                "cannot set the lower bound for BINARY or SPIN variables, "
                f"{v!r} is a {self.vartype(v).name} variable")

        if lb < cppvartype_info[bias_type].min(cppvartype):
            raise ValueError(f"lower_bound cannot be less than {cppvartype_info[bias_type].min(cppvartype)}")
            
        if lb > self.cppqm.upper_bound(vi):
            raise ValueError(
                f"the specified lower bound, {lb}, cannot be set greater than the "
                f"current upper bound, {self.cppqm.upper_bound(vi)}"
                )

        if cppvartype == cppVartype.INTEGER:
            if ceil(lb) > floor(self.cppqm.upper_bound(vi)):
                raise ValueError(
                    "there must be at least one integer value between "
                    f"the specified lower bound, {lb} and the "
                    f"current upper bound, {self.cppqm.upper_bound(vi)}"
                    )

        cdef bias_type *b = &(self.cppqm.lower_bound(vi))
        b[0] = lb

    def set_upper_bound(self, v, bias_type ub):
        cdef Py_ssize_t vi = self.variables.index(v)
        cdef cppVartype cppvartype = self.cppqm.vartype(vi)

        if cppvartype == cppVartype.BINARY or cppvartype == cppVartype.SPIN:
            raise ValueError(
                "cannot set the upper bound for BINARY or SPIN variables, "
                f"{v!r} is a {self.vartype(v).name} variable")

        if ub > cppvartype_info[bias_type].max(cppvartype):
            raise ValueError(f"upper_bound cannot be more than {cppvartype_info[bias_type].max(cppvartype)}")
            
        if ub < self.cppqm.lower_bound(vi):
            raise ValueError(
                f"the specified upper bound, {ub}, cannot be set less than the "
                f"current lower bound, {self.cppqm.lower_bound(vi)}"
                )

        if cppvartype == cppVartype.INTEGER:
            if ceil(self.cppqm.lower_bound(vi)) > floor(ub):
                raise ValueError(
                    "there must be at least one integer value between "
                    f"the specified upper bound, {ub} and the "
                    f"current lower bound, {self.cppqm.lower_bound(vi)}"
                    )

        cdef bias_type *b = &(self.cppqm.upper_bound(vi))
        b[0] = ub

    def set_quadratic(self, u, v, bias_type bias):
        cdef Py_ssize_t ui = self.variables.index(u)
        cdef Py_ssize_t vi = self.variables.index(v)

        if ui == vi and self.cppqm.vartype(ui) != cppVartype.INTEGER:
            raise ValueError(f"{u!r} cannot have an interaction with itself")
        
        self.cppqm.set_quadratic(ui, vi, bias)

    def upper_bound(self, v):
        cdef Py_ssize_t vi = self.variables.index(v)
        return as_numpy_float(self.cppqm.upper_bound(vi))

    def vartype(self, v):
        cdef Py_ssize_t vi = self.variables.index(v)
        cdef cppVartype cppvartype = self.cppqm.vartype(vi)

        if cppvartype == cppVartype.BINARY:
            return Vartype.BINARY
        elif cppvartype == cppVartype.SPIN:
            return Vartype.SPIN
        elif cppvartype == cppVartype.INTEGER:
            return Vartype.INTEGER
        else:
            raise RuntimeError("unexpected vartype")
